package MuGeN::Gui::MapDrawingWindow;

use strict;

use Gtk2;

use MuGeN::FeatureWidget;

use MuGeN::Gui::MapListWindow;
use MuGeN::Gui::DataSourceConfigWindow;
use MuGeN::Gui::ResultSourceConfigWindow;

use MuGeN::PaintDevice;
use MuGeN::PaintDevice::GtkPaintDevice;
use MuGeN::PaintDevice::PNGPaintDevice;
use MuGeN::PaintDevice::PSPaintDevice;
use MuGeN::PaintDevice::XFigPaintDevice;

use MuGeN::PhysicalMapWidget;

use MuGeN::CompAnalResultWidget;

use MuGeN::PreferencesHandler;

use MuGeN::ColorHandler::FileColorHandler;

our $VERSION='20050606';

sub set_preferences {
    my $mugen=$MuGeN::PreferencesHandler::Preferences->getDocumentElement();
    my @nodes=$mugen->getElementsByTagName('mapdisplay',0);
    my $mapdisplay=undef;
    if (scalar(@nodes)==0) {
	my $doctype=$MuGeN::PreferencesHandler::Preferences->getDoctype();
	$doctype->addElementDecl('mapdisplay','EMPTY');
	$doctype->addAttDef('mapdisplay','width','CDATA','#REQUIRED');
	$doctype->addAttDef('mapdisplay','bpl','CDATA','#REQUIRED');
	$doctype->addAttDef('mapdisplay','lines','CDATA','#REQUIRED');
	$doctype->addAttDef('mapdisplay','strands','(collapsed|expanded)','"expanded"');
	$doctype->addAttDef('mapdisplay','frames','(visible|hidden)','"visible"');
	
	$mapdisplay=$MuGeN::PreferencesHandler::Preferences->createElement('mapdisplay');
	$mapdisplay->setAttribute('width','800');
	$mapdisplay->setAttribute('bpl','5000');
	$mapdisplay->setAttribute('lines','4');
	$mapdisplay->setAttribute('strands','expanded');
	$mapdisplay->setAttribute('frames','visible');
	$mugen->appendChild($mapdisplay);
    } else {
	$mapdisplay=shift @nodes;
    }

    
    $MuGeN::Gui::MapDrawingWindow::width=$mapdisplay->getAttribute('width');
    $MuGeN::Gui::MapDrawingWindow::bpl=$mapdisplay->getAttribute('bpl');
    $MuGeN::Gui::MapDrawingWindow::lines=$mapdisplay->getAttribute('lines');
    $MuGeN::Gui::MapDrawingWindow::strands=$mapdisplay->getAttribute('strands');
    $MuGeN::Gui::MapDrawingWindow::frames=$mapdisplay->getAttribute('frames');

}

BEGIN {
    MuGeN::PreferencesHandler::register_listener(\&set_preferences);
}


my $MINBPL=50;
my $MINWIDTH=200;
my $ASPECT_RATIO=640/480;
my $MINHEIGHT=int($MINWIDTH/$ASPECT_RATIO);

sub new {
     my $class=shift;
     my $manager=shift;
     my ($width,$height)=@_;

     $width=$MuGeN::PhysicalMapWidget::width
	 if (!defined $width || $width<$MINWIDTH);
     $height=int($MuGeN::PhysicalMapWidget::width/$ASPECT_RATIO)
	 if (!defined $height || $height<$MINHEIGHT);

     my $self={};
     bless $self,$class;

     $self->{manager}=$manager;

     $self->{mapids}=();
     $self->{fwstartposition}=1;
     $self->{revstartposition}=1;
     $self->{nlines}=$MuGeN::PhysicalMapWidget::lines;
     $self->{bpl}=$MuGeN::PhysicalMapWidget::bpl;
     if ($MuGeN::PhysicalMapWidget::strands eq 'expanded') {
	 $self->{strands_expanded}=1;
	 MuGeN::ScalableFeatureWidget::expand_strands();
     } else {
	 $self->{strands_expanded}=0;
	 MuGeN::ScalableFeatureWidget::collapse_strands();
     }

     if ($MuGeN::PhysicalMapWidget::frames eq 'visible') {
	 $self->{frames_visible}=1;
	 MuGeN::ScalableFeatureWidget::show_frames();
     } else {
	 $self->{frames_visible}=0;
	 MuGeN::ScalableFeatureWidget::hide_frames();
     }


     $self->{width}=$width;
     $self->{height}=$height;

     my $window=Gtk2::Window->new('toplevel');
     $window->set_title('Multi-Genome Navigator ('.$VERSION.')');
     $self->{window}=$window;
     $window->set_resizable(1);
     $window->signal_connect('delete_event'=>sub { $manager->quit(); });
     $window->signal_connect('destroy_event'=>sub { $manager->quit(); });

     my $vbox=Gtk2::VBox->new(0,1);
     $vbox->show_all();
     $window->add($vbox);
     
     my $menubar=Gtk2::MenuBar->new;
     $vbox->pack_start($menubar,0,0,1);
     $self->{menubar}=$menubar;


     my $menuitem=Gtk2::MenuItem->new('File');
     $menubar->append($menuitem);
     my $submenu=Gtk2::Menu->new();
     $menuitem->set_submenu($submenu);
     $menuitem=Gtk2::MenuItem->new("Export Map");
     $submenu->append($menuitem);
     my $subsubmenu=Gtk2::Menu->new();
     $menuitem->set_submenu($subsubmenu);
     $menuitem =Gtk2::MenuItem->new("XFig");
     $menuitem->signal_connect("activate"=>sub{$self->_export_map('fig')});
     $subsubmenu->append($menuitem);
     $menuitem =Gtk2::MenuItem->new("PostScript");
     $subsubmenu->append($menuitem);
     my $pssubmenu=Gtk2::Menu->new();
     $menuitem->set_submenu($pssubmenu);
     foreach my $media (sort keys %MuGeN::PaintDevice::PAGE_WIDTHS_CM) {
	 if ($media ne 'custom') {
	     $menuitem=Gtk2::MenuItem->new($media);	 
	     $menuitem->signal_connect("activate"=>sub{$self->_export_map('ps',$media)});
	     $pssubmenu->append($menuitem);
	 }
     }
     $menuitem =Gtk2::MenuItem->new("Encapsulated PostScript");
     $menuitem->signal_connect("activate"=>sub{$self->_export_map('eps')});
     $subsubmenu->append($menuitem);     
     $menuitem =Gtk2::MenuItem->new("PNG");
     $menuitem->signal_connect("activate"=>sub{$self->_export_map('png')});
     $subsubmenu->append($menuitem);     
     $menuitem=Gtk2::MenuItem->new('Quit');
     $menuitem->signal_connect('activate'=>sub{$self->{manager}->quit();});
     $submenu->append($menuitem);

     $menuitem=Gtk2::MenuItem->new('Preferences');
     $menubar->append($menuitem);
     $submenu=Gtk2::Menu->new;
     $menuitem->set_submenu($submenu);
     $menuitem=Gtk2::CheckMenuItem->new('Expand Strands');
     if (!MuGeN::ScalableFeatureWidget::are_strands_collapsed) {
	 $menuitem->set_active(1);
	 $self->{strands_expanded}=1;
     }
     $menuitem->signal_connect('toggled'=>sub{$self->_change_strand_state;});
     $submenu->append($menuitem);
     $menuitem=Gtk2::CheckMenuItem->new('Show Frames');
     if (MuGeN::ScalableFeatureWidget::are_frames_visible) {
	 $menuitem->set_active(1);
	 $self->{frames_visible}=1;
     }
     $menuitem->signal_connect('toggled'=>sub{$self->_change_frame_state;});
     $submenu->append($menuitem);
     $menuitem=Gtk2::CheckMenuItem->new('Show Result Highlights');
     if ($MuGeN::CompAnalResultWidget::show_highlights) {
	 $menuitem->set_active(1);
	 $self->{highlights_visible}=1;
     }
     $menuitem->signal_connect('toggled'=>sub{$self->_change_highlight_state;});     $submenu->append($menuitem);
     $menuitem=Gtk2::MenuItem->new('Visible Features');
     $submenu->append($menuitem);
     $self->{featureitem}=$menuitem;

     $menuitem=Gtk2::MenuItem->new('Map Area Width');
     $submenu->append($menuitem);
     $subsubmenu=Gtk2::Menu->new();
     $menuitem->set_submenu($subsubmenu);
     my $previtem=undef;
     my %sizes=(320=> 1,
		640=> 1,
		800=> 1,
		1024=>1,
		1050=>1,
		1152=>1,
		1280=>1,
		1400=>1,
		1600=>1,
		1920=>1,
		2000=>1,
		2400=>1,
		3000=>1,
		3600=>1,
		4000=>1,
		4800=>1,
		5600=>1,
		6400=>1,
		7200=>1,
		8000=>1,
		9000=>1,
		10000=>1);
     $sizes{$width}=1;
     foreach my $size (sort { $a <=> $b } keys %sizes) {
	 if (defined $previtem) {
	     $menuitem=Gtk2::RadioMenuItem->new($previtem,"$size");
	 } else {
	     $menuitem=Gtk2::RadioMenuItem->new("$size");

	 }
	 $menuitem->set_active(1)
	     if ($width == $size);
	 $menuitem->signal_connect('activate'=>sub{$self->_set_width($size);});
	 $subsubmenu->append($menuitem);
	 $previtem=$menuitem;
     }
     

     $menuitem=Gtk2::MenuItem->new('Colormaps');
     $submenu->append($menuitem);
     $subsubmenu=Gtk2::Menu->new();
     $menuitem->set_submenu($subsubmenu);

     $menuitem=Gtk2::MenuItem->new('Load Colormap...');
     $menuitem->signal_connect('activate'=>sub{$self->_load_colormap;});
     $subsubmenu->append($menuitem);

     $menuitem=Gtk2::MenuItem->new('Reset Colormap');
     $menuitem->signal_connect('activate'=>sub{$self->_reset_colormap;});
     $subsubmenu->append($menuitem);


     my $dscw=MuGeN::Gui::DataSourceConfigWindow->new($self->{manager});
     $self->{datasourceconfigwindow}=$dscw;
     $menuitem=Gtk2::MenuItem->new('Configure Data Sources...');
     $menuitem->signal_connect('activate'=>sub{$self->{datasourceconfigwindow}->show()});
     $submenu->append($menuitem);


     my $rscw=MuGeN::Gui::ResultSourceConfigWindow->new($self->{manager});
     $self->{resultsourceconfigwindow}=$rscw;
     $menuitem=Gtk2::MenuItem->new('Configure Result Sources...');
     $menuitem->signal_connect('activate'=>sub{$self->{resultsourceconfigwindow}->show()});
     $submenu->append($menuitem);


     $menuitem=Gtk2::MenuItem->new('Save Preferences...');
     $menuitem->signal_connect('activate'=>sub{$self->_save_preferences;});
     $submenu->append($menuitem);

     $menuitem=Gtk2::MenuItem->new('Windows');
     $menubar->append($menuitem);
     $submenu=Gtk2::Menu->new();
     $menuitem->set_submenu($submenu);
     $menuitem=Gtk2::CheckMenuItem->new('Map List Window');
     $menuitem->set_active(1);
     $menuitem->signal_connect('toggled' => sub { $self->{manager}->change_window_visibility('listwindow',0); return 1});
     $self->{listwindowitem}=$menuitem;
     $submenu->append($menuitem);
     $menuitem=Gtk2::CheckMenuItem->new('Highlights/Feature Window');
     $menuitem->set_active(1);
     $menuitem->signal_connect('toggled' => sub { $self->{manager}->change_window_visibility('featurewindow',0); return 1});
     $self->{featurewindowitem}=$menuitem;
     $submenu->append($menuitem);
     $menuitem=Gtk2::CheckMenuItem->new('Result Information Window');
     $menuitem->set_active(1);
     $menuitem->signal_connect('toggled' => sub { $self->{manager}->change_window_visibility('resultlogwindow',0); return 1});
     $self->{resultlogwindowitem}=$menuitem;
     $submenu->append($menuitem);

     my $frame=Gtk2::Frame->new();
     $self->{frame}=$frame;
     $vbox->pack_start($frame,1,1,1);

     my $paintdevice=MuGeN::PaintDevice::GtkPaintDevice->new($width,$height,$frame);
     $self->{paintdevice}=$paintdevice;

     my $mapwidget=new MuGeN::PhysicalMapWidget($paintdevice);
     $self->{mapwidget}=$mapwidget;

     my $hbox=Gtk2::HBox->new(0,1);
     $hbox->show();
     $vbox->pack_start($hbox,0,0,1);

     my $arrowbox=Gtk2::HBox->new(0,0);
     my $alignment=Gtk2::Alignment->new(1.0,0.5,0.0,1.0);
     $arrowbox->add($alignment);
     my $arrow=Gtk2::Arrow->new('left','out');
     $alignment->add($arrow);
     $alignment=Gtk2::Alignment->new(0.0,0.5,0.0,1.0);
     $arrowbox->add($alignment);
     $arrow=Gtk2::Arrow->new('left','out');
     $alignment->add($arrow);

     my $prevbtn=Gtk2::Button->new;
     $prevbtn->add($arrowbox);
     $prevbtn->show();
     $prevbtn->signal_connect('clicked',sub {$self->_move_btn_clicked(-1);});
     $hbox->pack_start($prevbtn,1,1,0);

     $arrow=Gtk2::Arrow->new('left','out');
     $prevbtn=Gtk2::Button->new;
     $prevbtn->add($arrow);
     $prevbtn->show();
     $prevbtn->signal_connect('clicked',sub {$self->_move_btn_clicked(-0.5);});
     $hbox->pack_start($prevbtn,1,1,2);

     $arrow=Gtk2::Arrow->new('right','out');
     my  $nextbtn=Gtk2::Button->new();
     $nextbtn->add($arrow);
     $nextbtn->show();
     $nextbtn->signal_connect('clicked',sub {$self->_move_btn_clicked(0.5);});
     $hbox->pack_start($nextbtn,1,1,1);

     $arrowbox=Gtk2::HBox->new(0,0);
     $alignment=Gtk2::Alignment->new(1.0,0.5,0.0,1.0);
     $arrowbox->add($alignment);
     $arrow=Gtk2::Arrow->new('right','out');
     $alignment->add($arrow);
     $alignment=Gtk2::Alignment->new(0.0,0.5,0.0,1.0);
     $arrowbox->add($alignment);
     $arrow=Gtk2::Arrow->new('right','out');
     $alignment->add($arrow);

     $nextbtn=Gtk2::Button->new();
     $nextbtn->add($arrowbox);
     $nextbtn->show();
     $nextbtn->signal_connect('clicked',sub {$self->_move_btn_clicked(1);});
     $hbox->pack_start($nextbtn,1,1,1);

     $hbox=Gtk2::HBox->new(0,1);
     $hbox->show();
     $vbox->pack_start($hbox,0,0,1);

     my $zoomoutoutbtn=Gtk2::Button->new('--');
     $zoomoutoutbtn->show();
     $zoomoutoutbtn->signal_connect('clicked',
				    sub {$self->_zoom_btn_clicked(4);});
     $hbox->pack_start($zoomoutoutbtn,1,1,0);

     my $zoomoutbtn=Gtk2::Button->new('-');
     $zoomoutbtn->show();
     $zoomoutbtn->signal_connect('clicked',
				 sub {$self->_zoom_btn_clicked(2);});
     $hbox->pack_start($zoomoutbtn,1,1,0);

     my $zoominbtn=Gtk2::Button->new('+');
     $zoominbtn->show();
     $zoominbtn->signal_connect('clicked',
				sub {$self->_zoom_btn_clicked(0.5);});
     $hbox->pack_start($zoominbtn,1,1,0);

     my $zoomininbtn=Gtk2::Button->new('++');
     $zoomininbtn->show();
     $zoomininbtn->signal_connect('clicked',
				  sub {$self->_zoom_btn_clicked(0.25);});
     $hbox->pack_start($zoomininbtn,1,1,0);

     $hbox=Gtk2::HBox->new(0,4);
     $hbox->show();
     $vbox->pack_start($hbox,0,0,1);

     my $label=Gtk2::Label->new('Start base');
     $hbox->pack_start($label,0,1,1);
     my $entry=Gtk2::Entry->new_with_max_length(9);
     
     $hbox->pack_start($entry,1,1,1);
     $entry->signal_connect('activate',sub {$self->_map_entry_changed();});
     $entry->set_text($self->{fwstartposition});
     $self->{startentry}=$entry;
     $label=Gtk2::Label->new('Lines');
     $hbox->pack_start($label,0,1,1);
     my $bpladj=Gtk2::Adjustment->new($self->{nlines},1,20,1,2,2);
     $self->{bpladj}=$bpladj;
     my $spinbtn=Gtk2::SpinButton->new($bpladj,1,0);
     $hbox->pack_start($spinbtn,1,1,1);
     $spinbtn->set_text($self->{nlines});
     $self->{linesentry}=$spinbtn;
     $bpladj->signal_connect('value_changed', sub {$self->_map_entry_changed();});
     $spinbtn->signal_connect('activate',sub {$self->_map_entry_changed();});
     $label=Gtk2::Label->new('BPL');
     $hbox->pack_start($label,0,1,1);
     $entry=Gtk2::Entry->new_with_max_length(9);
     $entry->signal_connect('activate',sub {$self->_map_entry_changed();});
     $hbox->pack_start($entry,1,1,1);
     $entry->set_text($self->{bpl});
     $self->{bplentry}=$entry;


     $hbox=Gtk2::HBox->new(0,4);
     $hbox->show();
     $vbox->pack_start($hbox,0,0,1);

     $label=Gtk2::Label->new('Birds eye view thresh. (kbp) :');
     $hbox->pack_start($label,0,1,1);
     my $viewthreshadj=new Gtk2::Adjustment($mapwidget->get_birdseye_display_threshold/1000,10,1001,1,1,1);
     $viewthreshadj->signal_connect('value_changed',sub {$self->_map_entry_changed();});
     $self->{viewthreshadj}=$viewthreshadj;
     my $scale=Gtk2::HScale->new($viewthreshadj);
     $scale->set_update_policy('delayed');
     $scale->set_draw_value(1);
     $scale->set_digits(0);
     $hbox->pack_start($scale,1,1,1);
     
     $label=Gtk2::Label->new('Seq. display thresh. (bp) :');
     $hbox->pack_start($label,0,1,1);
     my $seqthreshadj=Gtk2::Adjustment->new($mapwidget->get_sequence_display_threshold,1,1001,1,1,1);
     $seqthreshadj->signal_connect('value_changed',sub {$self->_map_entry_changed();});
     $self->{seqthreshadj}=$seqthreshadj;
     $scale=Gtk2::HScale->new($seqthreshadj);
     $scale->set_update_policy('delayed');
     $scale->set_draw_value(1);
     $scale->set_digits(0);
     $hbox->pack_start($scale,1,1,1);
     
     $label=Gtk2::Label->new('Name display thresh.');
     $hbox->pack_start($label,0,1,1);
     my $namethreshadj=Gtk2::Adjustment->new($mapwidget->get_text_display_threshold,0.0,1.1,0.01,0.01,0.1);
     $namethreshadj->signal_connect('value_changed',sub {$self->_map_entry_changed();});
     $self->{namethreshadj}=$namethreshadj;
     $scale=Gtk2::HScale->new($namethreshadj);
     $scale->set_update_policy('delayed');
     $scale->set_draw_value(1);
     $scale->set_digits(3);
     $hbox->pack_start($scale,1,1,1);


     MuGeN::PreferencesHandler::save_preferences();
     

     return $self;
}

sub append_map {
    my $self=shift;
    my $map=shift;

    my $start=$self->{fwstartposition};
    my $nlines=$self->{nlines};
    my $bpl=$self->{bpl};
    my $end=$start+$nlines*$bpl-1;

    my $mapid=$self->{mapwidget}->append_map($map,$start,$end,$bpl);

    push @{$self->{mapids}},$mapid;

    $self->{mapwidget}->render();
    $self->_rebuild_feature_menu;

    return $mapid;
}

sub shift_map_up {
    my $self=shift;
    my $mapid=shift;

    $self->{mapwidget}->shift_map_up($mapid);
    $self->{mapwidget}->render();
}

sub shift_map_down {
    my $self=shift;
    my $mapid=shift;

    $self->{mapwidget}->shift_map_down($mapid);
    $self->{mapwidget}->render();
}


sub hide_map {
    my $self=shift;
    my $mapid=shift;

    $self->{mapwidget}->hide_map($mapid);
    $self->{mapwidget}->render();
}

sub show_map {
    my $self=shift;
    my $mapid=shift;

    $self->{mapwidget}->show_map($mapid);
    $self->{mapwidget}->render();
}



sub replace_map {
    my $self=shift;
    my $repmapid=shift;
    my $newmap=shift;

    $self->{mapwidget}->replace_map($newmap,$repmapid);
    $self->{mapwidget}->render();
}

sub remove_map {
    my $self=shift;
    my $delmapid=shift;

    my $newmapids=[];
    foreach my $mapid (@{$self->{mapids}}) {
	push @{$newmapids},$mapid
	    if ($mapid != $delmapid);
    }

    $self->{mapids}=$newmapids;
    $self->{mapwidget}->remove_map($delmapid);
    $self->{mapwidget}->render();
}

sub render {
    my $self=shift;
    $self->{mapwidget}->render();
}


sub add_listener {
    my $self=shift;
    my $listener=shift;
    my $type=shift;

    $self->{paintdevice}->add_listener($listener,$type);
}

sub show_all {
    my $self=shift;
    $self->{window}->show_all();
}

sub get_start_position {
    my $self=shift;
    return $self->{fwstartposition};
}

sub set_start_position {
    my $self=shift;
    my $start=shift;

    my $deltastart=$start-$self->{fwstartposition};
    $self->{fwstartposition}=$start;
    $self->{revstartposition}=$self->{revstartposition}-$deltastart;
    $self->{startentry}->set_text($start);

    $self->update_map_display;
}

sub get_bpl {
    my $self=shift;
    return $self->{bpl};
}

sub set_bpl {
    my $self=shift;
    my $bpl=shift;
    
    $self->{bpl}=$bpl;
    $self->{bplentry}->set_text($bpl);
    $self->update_map_display;
}


sub get_display_parameters {
    my $self=shift;
    
    my $endposition=$self->{fwstartposition}+$self->{nlines}*$self->{bpl}-1;
    return ($self->{fwstartposition},$endposition,$self->{bpl});
}

sub set_display_parameters {
    my $self=shift;
    my ($start,$end,$bpl)=@_;
    
    my $deltastart=$start-$self->{fwstartposition};
    $self->{fwstartposition}=$start;
    $self->{bpl}=$bpl;
    $self->{nlines}=($end-$start+1)/$bpl;
    $self->{nlines}=1
	if ($self->{nlines}<1);
    $self->{revstartposition}=$self->{revstartposition}-$deltastart;

    $self->{startentry}->set_text($start);
    $self->{linesentry}->set_text($self->{nlines});
    $self->{bplentry}->set_text($bpl);
    $self->update_map_display();
}


sub update_map_display {
    my $self=shift;

    my $fwstart=$self->{fwstartposition};
    my $revstart=$self->{revstartposition};
    my $nlines=$self->{nlines};
    my $bpl=$self->{bpl};

    if ($fwstart==0) {
	$fwstart++;
	$self->{fwstartposition}=$fwstart;
	$self->{startentry}->set_text($fwstart);
    }
    if ($fwstart+$nlines*$bpl-1==0) {
	$fwstart--;
	$self->{fwstartposition}=$fwstart;
	$self->{startentry}->set_text($fwstart);
	
    }
    
    if ($revstart==0) {
	$revstart++;
	$self->{revstartposition}=$revstart;
    }
    if ($revstart+$nlines*$bpl-1==0) {
	$revstart--;
	$self->{revstartposition}=$revstart;
	
    }
    
    foreach my $mapid (@{$self->{mapids}}) {
	my $anchor=$self->{manager}->get_anchor_position('ID'=>$mapid);
	my $mapstart;
	my $mapend;
	if (! $self->{manager}->is_reverse_complemented($mapid)) {
	    $mapstart=$fwstart+$anchor-1;
	    $mapend=$mapstart+$nlines*$bpl-1;
	} else {
	    $mapstart=$revstart+$anchor-1;
	    $mapend=$mapstart-$nlines*$bpl+1;
	}

	$self->{mapwidget}->set_view_range($mapid,$mapstart,$mapend,$bpl);
    }
    $self->{mapwidget}->render();
    $self->_rebuild_feature_menu;
}

sub change_window_visibility {
    my $self=shift;
    my $windowname=shift;

    if ($self->{$windowname."item"}->get_active) {
	$self->{$windowname."item"}->set_active(0);
    } else {
	$self->{$windowname."item"}->set_active(1);
    }

}

sub _set_width {
    my $self=shift;

    return 1
	if (!defined $self->{frame});
    my $width=shift;
    $self->{width}=$width;
    my $height=$self->{height};
    my $paintdevice=$self->{mapwidget}->get_paintdevice();
    $paintdevice->resize($width,$height);
    $self->{mapwidget}->render();

    my @nodes=$MuGeN::PreferencesHandler::Preferences->getElementsByTagName('mapdisplay',1);
    my $mapdisplay=shift @nodes;
    $mapdisplay->setAttribute('width',$width);
    
    return 1;
}

sub _move_btn_clicked {
    my $self=shift;
    my $direction=shift;

    my $fwstart=$self->{fwstartposition};
    my $revstart=$self->{revstartposition};
    my $nlines=$self->{nlines};
    my $bpl=$self->{bpl};
    my $newfwstart=$fwstart+$direction*$bpl*$nlines;
    my $newrevstart=$revstart-$direction*$bpl*$nlines;
    $self->{fwstartposition}=$newfwstart;
    $self->{revstartposition}=$newrevstart;
    $self->update_map_display;
    $newfwstart=$self->{fwstartposition};
    $self->{startentry}->set_text($newfwstart);

    return 1;
}

sub _zoom_btn_clicked {
    my $self=shift;
    my $factor=shift;

    my $fwstart=$self->{fwstartposition};
    my $revstart=$self->{revstartposition};
    my $nlines=$self->{nlines};
    my $bpl=$self->{bpl};
    my $range=$bpl*$nlines;    
    $range=$range*$factor;
    my $newbpl=int($range/$nlines);
    $self->{bpl}=$newbpl;
    $self->{bplentry}->set_text($newbpl);
    $self->update_map_display;

    return 1;
}

sub _rebuild_feature_menu {
    my $self=shift;

    my @featurenames=MuGeN::FeatureWidgetFactory::get_feature_names();

    if (defined $self->{featuremenu}) {
	$self->{featureitem}->remove_submenu();
	$self->{featuremenu}->destroy();
    }

    $self->{featuremenu}=Gtk2::Menu->new();

    foreach my $name (@featurenames) {
	my $item=Gtk2::CheckMenuItem->new($name);
	if (MuGeN::FeatureWidgetFactory::is_masked($name)) {
	    $item->set_active(0);
	} else {
	    $item->set_active(1);
	}
	$item->signal_connect("toggled"=>sub{$self->_toggle_feature_mask($name);});
	$self->{featuremenu}->append($item);
    }
    
    $self->{featureitem}->set_submenu($self->{featuremenu});
    $self->{menubar}->show_all();
}


sub _toggle_feature_mask {
    my $self=shift;
    my $name=shift;
    
    if (MuGeN::FeatureWidgetFactory::is_masked($name)) {
      MuGeN::FeatureWidgetFactory::unmask_feature($name);
    } else {
      MuGeN::FeatureWidgetFactory::mask_feature($name);
    }
    $self->{mapwidget}->render;

    return 1;
}

sub _change_strand_state {
    my $self=shift;

    my $prefstring='';
    if ($self->{strands_expanded}) {
      MuGeN::ScalableFeatureWidget::collapse_strands();
	$self->{strands_expanded}=0;
	$prefstring='collapsed';
    } else {
      MuGeN::ScalableFeatureWidget::expand_strands();
	$self->{strands_expanded}=1;
	$prefstring='expanded';
    }

    my @nodes=$MuGeN::PreferencesHandler::Preferences->getElementsByTagName('mapdisplay');
    my $mapdisplay=shift @nodes;
    $mapdisplay->setAttribute('strands',$prefstring);
    $self->{mapwidget}->render;

    return 1;
}

sub _change_frame_state {
    my $self=shift;

    my $prefstring='';
    if ($self->{frames_visible}) {
      MuGeN::ScalableFeatureWidget::hide_frames();
	$self->{frames_visible}=0;
	$prefstring='hidden';
    } else {
      MuGeN::ScalableFeatureWidget::show_frames();
	$self->{frames_visible}=1;
	$prefstring='visible';
    }

    my @nodes=$MuGeN::PreferencesHandler::Preferences->getElementsByTagName('mapdisplay');
    my $mapdisplay=shift @nodes;
    $mapdisplay->setAttribute('frames',$prefstring);
    $self->{mapwidget}->render;

    return 1;
}


sub _change_highlight_state {
    my $self=shift;

    my @nodes=$MuGeN::PreferencesHandler::Preferences->getElementsByTagName('highlights');
    my $highlights=shift @nodes;
    if ($self->{highlights_visible}) {
	$MuGeN::CompAnalResultWidget::show_highlights=0;
	$self->{highlights_visible}=0;
	$highlights->setAttribute('visible','false');
    } else {
	$MuGeN::CompAnalResultWidget::show_highlights=1;
	$self->{highlights_visible}=1;
	$highlights->setAttribute('visible','true');
    }
    $self->{mapwidget}->render;

    return 1;
}

sub _export_map {
    my $self=shift;
    my $format = shift;
    my $media=shift;

    my $dialog =Gtk2::FileSelection->new("Export Map");
    $dialog->set_filename("map.$format");
    $dialog->ok_button->signal_connect("clicked",sub {$self->_write_map_file($dialog,$format,$media)});
    $dialog->ok_button->signal_connect("clicked",sub {$dialog->destroy(); 
						      return 1});
    $dialog->cancel_button->signal_connect("clicked",sub {$dialog->destroy();
						      return 1});
    
    $dialog->show();

    return 1;
}

sub _write_map_file {
    my $self=shift;
    my $dialog = shift;
    my $format = shift;

    my $filename=$dialog->get_filename;
    my $filepaintdevice;
    my $oldpaintdevice=$self->{mapwidget}->get_paintdevice;
    my $screenresolution=$MuGeN::PaintDevice::GtkPaintDevice::SCREEN_RESOLUTION;
    
    if ($format eq 'ps') {
	my $media=shift;
	my $pagewidth=$MuGeN::PaintDevice::PAGE_WIDTHS_CM{$media};
	my $pageheight=$MuGeN::PaintDevice::PAGE_HEIGHTS_CM{$media};
	$filepaintdevice=new MuGeN::PaintDevice::PSPaintDevice($filename,$pagewidth,
							$pageheight,
							type=>'ps',
							media=>$media);
    }
    if ($format eq 'eps') {
	my $width_cm=$self->{width}/$screenresolution*$MuGeN::PaintDevice::PSPaintDevice::CM_PER_INCH;
	my $height_cm=$width_cm*$self->{height}/$self->{width};
	$filepaintdevice=new MuGeN::PaintDevice::PSPaintDevice($filename,$width_cm,
							$height_cm,
							type=>'eps');
    }


    if ($format eq 'fig') {
	my $width_inch=$self->{width}/$screenresolution;
	my $height_inch=$width_inch*$self->{height}/$self->{width};
	$filepaintdevice=new MuGeN::PaintDevice::XFigPaintDevice($filename,
							  $width_inch,
							  $height_inch);
    }
	

    my $factor=MuGeN::PaintDevice::PNGPaintDevice->text_height/$oldpaintdevice->text_height;
    $filepaintdevice=new MuGeN::PaintDevice::PNGPaintDevice($filename,$oldpaintdevice->width,int($oldpaintdevice->height*$factor))
	if ($format eq 'png');

    $self->{mapwidget}->set_paintdevice($filepaintdevice);
    $self->{mapwidget}->render;

    $self->{mapwidget}->set_paintdevice($oldpaintdevice);

    return 1;
}

sub _map_entry_changed {
    my $self=shift;

    my $do_update=0;

    my @nodes=$MuGeN::PreferencesHandler::Preferences->getElementsByTagName('mapdisplay');
    my $mapdisplay=shift @nodes;

    my $newstart=$self->{startentry}->get_text();
    if ( !($newstart =~ /-?\d+/) || $newstart==0) {
	
	$self->{startentry}->set_text($self->{fwstartposition});
    } else {
	my $deltastart=$newstart-$self->{fwstartposition};
        $self->{revstartposition}=$self->{revstartposition}-$deltastart;
 	$self->{fwstartposition}=$newstart;
	$do_update=1;
    }
    my $newnlines=$self->{linesentry}->get_text();
    if ($newnlines =~ /\D/ || $newnlines < 1) {
	$self->{linesentry}->set_text($self->{nlines});
    } else {
	$self->{nlines}=$newnlines;
	$mapdisplay->setAttribute('lines',$newnlines);
	$do_update=1;
    }
    my $newbpl=$self->{bplentry}->get_text();
    if ($newbpl =~ /\D/ || $newbpl < $MINBPL) {
	$self->{bplentry}->set_text($self->{bpl});
    } else {
	$self->{bpl}=$newbpl;
	$mapdisplay->setAttribute('bpl',$newbpl);
	$do_update=1;
    }

    my $oldthresh=$MuGeN::ScalableFeatureWidget::display_threshold;
    my $newthresh=$self->{namethreshadj}->value;
    if ($newthresh != $oldthresh) {
	$self->{mapwidget}->set_text_display_threshold($newthresh);
	$do_update=1;
    }

    my $oldseqthresh=$self->{mapwidget}->get_sequence_display_threshold();
    my $newseqthresh=$self->{seqthreshadj}->value();
    if ($newseqthresh != $oldseqthresh) {
	$self->{mapwidget}->set_sequence_display_threshold($newseqthresh);
	$do_update=1;
    }

    my $oldviewthresh=$self->{mapwidget}->get_birdseye_display_threshold();
    my $newviewthresh=$self->{viewthreshadj}->value();
    if ($newviewthresh != $oldviewthresh) {
	$self->{mapwidget}->set_birdseye_display_threshold($newviewthresh*1000);
	$do_update=1;
    }


    $self->update_map_display
	if ($do_update);

    return 1;
}

sub _save_preferences {
    my $self=shift;

    MuGeN::PreferencesHandler::save_preferences;

    return 1;

}

sub _load_colormap {
    my $self=shift;
    
    my $dialog=new Gtk2::FileSelection('Load Colormap');
    
    $dialog->ok_button->signal_connect('clicked',sub {$dialog->hide(); $self->_colormap_selected($dialog);});
    $dialog->cancel_button->signal_connect('clicked',sub {$dialog->hide(); return 1;});

    $dialog->show();

    return 1;
}

sub _reset_colormap {
    my $self=shift;

    $MuGeN::FeatureWidget::MuGeN::ColorHandler=new MuGeN::ColorHandler::DefaultColorHandler;
    $self->{mapwidget}->render;

}

sub _colormap_selected {
    my $self=shift;
    my $dialog=shift;

    my $file=$dialog->get_filename();

    $MuGeN::FeatureWidget::ColorHandler=new MuGeN::ColorHandler::FileColorHandler;
    $MuGeN::FeatureWidget::ColorHandler->load_colors($file);    

    $self->{mapwidget}->render;
}

sub destroy {
    my $self=shift;
    
    $self->{window}->destroy();
}


1
